import os
import re
import json
import ijson
import asyncio
from tqdm import tqdm
from . import utils
from . import params

class DeduplicateClient : 
    """
    This class allows to deduplicate a set of records
    Initialize params :
    :param fpath : filepath of the set of records 
    :param es : an instance of elasticsearc client
    :param index : index of elasticsearch database
    """

    def __init__(self, fpath, es, index) : 
        self.fpath = fpath
        self.es = es
        self.index = index

    async def one_iteration(self, id, item) : 
        """
        Asynchronous version of deduplicate function
        :param id : identifiant
        item : record
        :return a dictionary : 
            - id the id param
            - duplicate : list of duplicate found 
        """
        d = RecordDeduplicate(item, self.es, self.index)
        duplicate = d.deduplicate()
        data = {"id" : id, "duplicates" : duplicate}
        return data

    def save_file(fname, res) : 
        """
        save file at json format
        :param : fname : filename
        :res : a dictionary
        """
        with open(fname, "w", encoding="utf8") as f : 
            json.dump(res, f)

    async def main(self) : 
        """
        Main is an asynchrosous function which allows 
        
        """
        tasks = []
        with open(self.fpath) as f : 
            for i, item in tqdm(enumerate(ijson.items(f, "item"))) : 
                tasks.append(asyncio.ensure_future(self.one_iteration(i, item)))
        features = await asyncio.gather(*tasks)
        return features


    async def async_deduplicate(self,record) : 
        """
        Build an async deduplicate function 
        :param record : bibliographic record of type dict
        :return a dictionnary (see duplicate function below for details)
        """
        d = RecordDeduplicate(record, es = self.es)
        res = d.deduplicate()
        return res

class RecordDeduplicate :
    def __init__(self, record, es, index) :
        self.record = record
        self.es = es
        self.index = index
        self.dupList = []
        self.result_field = ["deduplicateRules", "idConditor", "source", "sourceUid"]

    def preprocess(self, sent) :
        """
        split a sentence, remove stopwords and join token
        :param sentence : a sequence of tokens
        :return preprocessed sentence : sentence where stop word is removed 
        """
        if not isinstance(sent, str) : 
            sent = str(sent)
        #lang = detect(sent)
        #tokens = ([fr.stem(word) for word in sent.split()],[en.stem(word) for word in sent.split()])[lang == "fr"]
        tokens = [token for token in re.findall("\w+", sent) \
            if token not in params.titleStopwords]
        return " ".join(tokens)

    def query(self) :
        """
        For a given sentence, preprocess and query elastic
        :param sentence
        :return elasticsearch responses
        see https://www.elastic.co/guide/en/elasticsearch/reference/current/search-your-data.html
        for details.
        """
        titleDefault = self.record['title']['default']
        req = self.preprocess(titleDefault)
        body = {
            'query': {
                'bool': {
                    'must': [
                        { 'match': {
                            'title.default': {
                            'query' : req,
                            'fuzziness': "AUTO",
                            'minimum_should_match': '60%'
                                }
                            }
                        }]
                        }
                    }
                }

        res = self.es.search(index = self.index, body = body)
        return res


    def deduplicate(self) :
        """
        Deduplicate function.
        :param record : bibliographic record
        :return duplicates if exists. duplicates are list of dict. each dict contains : 
         - idConditor
         - source
         - sourceUid
         - decision
         - deduplicateRules        
        """
        if not isinstance(self.record, dict) :
            return {"error" : {"code" : 104, "type" : "<class 'Type'>", "message" : "Must be an Object"}}

        if not bool(self.record):
            return {"error" : {"code" : 103, "type" : "<class 'ValueError'>", "message" : "Empty object"}}

        if "sourceUid" not in self.record :
            return {"error" : {"code" : 101, "type" : "<class 'KeyError'>", "message" : "KeyError('sourceUid')"}}

        # take idConditor for duplicates if exists
        duplicatesIdConditor = []
        try :
            duplicatesIdConditor = [x["idConditor"] for x in self.record["duplicates"]]
        except Exception as err :
            pass

        if "idConditor" in self.record :
            duplicatesIdConditor.append(self.record["idConditor"])

        res = self.query()

        if res["hits"] :
            for rc in res["hits"]["hits"] :
                dupCandidateRecord = rc["_source"]
                comp = utils.RecordComparison(self.record, dupCandidateRecord)
                comp.run()
                if comp.result == 1 :
                    dico = {}
                    for field in self.result_field :
                        try :
                            dico[field] = dupCandidateRecord[field]
                        except : pass

                    dico["deduplicateRules"] = comp.duplicate_rules
                    dico['type'] = dupCandidateRecord["typeConditor"]
                    self.dupList.append(dico)

            return [duplicate for duplicate in self.dupList if duplicate["sourceUid"] != self.record["sourceUid"]]
        else : return []