from .params import decisionGrid, typeConditorCategory
from .utils.tools import getSettlement
from .utils.pairwise import *

class ConditorRecord :

    def __init__(self, record) :
        assert isinstance(record, dict), "Input must be a dict"
        # id
        self.doi = record["doi"] if "doi" in record else None
        self.pmId = record["pmId"] if "pmId" in record else None
        self.nnt = record["nnt"] if "nnt" in record else None

        # Content
        self.titleDefault = record["title"]['default'] if "title" in record else None
        self.en = record["title"]["en"] if "title" in record else None
        self.fr = record["title"]["fr"] if "title" in record else None
        self.titleDict = {"default" : self.titleDefault, "en" : self.en, "fr" : self.fr}

        # container
        self.meeting = record["title"]["meeting"] if "title" in record else None
        self.journal = record["title"]["journal"] if "title" in record else None
        self.issn = record["issn"] if "issn" in record else None
        self.eissn = record["eissn"] if "eissn" in record else None
        self.eisbn = record["eisbn"] if "eisbn" in record else None
        self.isbn = record["isbn"] if "isbn" in record else None
        self.monography = record["title"]["monography"] if "title" in record else None
        self.source = record["source"] if "source" in record else None
        self.docType = record["documentType"][0] if "documentType" in record else None
        self.publiSource = getSettlement(record["teiBlob"]) if "teiBlob" in record else None

        # Positionning
        self.publiDate = record["publicationDate"] if "publicationDate" in record else None
        self.pageRange = record["pageRange"] if "pageRange" in record else None
        self.issue = record["issue"] if "issue" in record else None
        self.volume = record["volume"] if "volume" in record else None

        # Others
        self.typeConditor = record["typeConditor"] if "typeConditor" in record else None
        self.idConditor = record["idConditor"] if "idConditor" in record else None
        self.sourceUid = record['sourceUid'] if "sourceUid" in record else None

#########################################
#### Field extraction for corhal ########
#########################################

class CorhalRecord :

    def __init__(self, record) :
        assert isinstance(record, dict), "Input must be a dict"

        # id
        self.doi = record["doi"] if "doi" in record else None
        self.pmId = record["pmId"] if "pmId" in record else None
        self.nnt = record["nnt"] if "nnt" in record else None

        # Content
        self.titleDefault = record["title"]['default'] if "title" in record else None
        self.en = record["title"]["en"] if "title" in record else None
        self.fr = record["title"]["fr"] if "title" in record else None
        self.titleDict = {"default" : self.titleDefault, "en" : self.en, "fr" : self.fr}

        # container
        self.meeting = record["host"]["conference"]["name"] if "host" in record else None
        self.journal = record["host"]["title"] if "host" in record else None
        self.issn = record["host"]["issn"] if "host" in record else None
        self.eissn = record["host"]["eissn"] if "host" in record else None
        self.eisbn = record["host"]["eisbn"] if "host" in record else None
        self.isbn = record["host"]["isbn"] if "host" in record else None
        self.monography = "" #record["title"]["monography"] if "title" in record else None
        self.source = record["source"] if "source" in record else None
        self.originalGenre = record["originalGenre"] if "originalGenre" in record else None # change
        self.publiSource = getSettlement(record["teiBlob"]) if "teiBlob" in record else None

        # Positionning
        self.publiDate = record["_business"]["xPublicationDate"] if "_business" in record else None
        self.pageRange = record["host"]["pages"][0]["range"] if "host" in record else None
        self.issue = record["host"]["issue"] if "host" in record else None
        self.volume = record["host"]["volume"] if "host" in record else None

        # Others
        self.duplicateGenre = record["_business"]["duplicateGenre"] if "_business" in record else None # change
        self.idIstex = record["idIstex"] if "idIstex" in record else None # change
        self.sourceUid = record['sourceUid'] if "sourceUid" in record else None


#########################################
#### features comparison for Conditor ###
#########################################

class ConditorRecordComparison :
    """
    Compare two records with a threshold for title distance 
    (levenshtein distance)
    """

    def __init__(self, record1, record2, threshold = 0.2) :

        self.n1 = self.Record(record1)
        self.n2 = self.Record(record2)

        self.threshold = threshold
        self.validation_dict = {}
        self.result = None
        self.duplicate_rules = None
        self.status = False

    def compareId(self):
       self.validation_dict["id"] = compareId(self.n1.doi, self.n2.doi,\
                           self.n1.nnt, self.n2.nnt,\
                           self.n1.pmId, self.n2.pmId)

    def comparePageRange(self) :
       self.validation_dict["page"] = comparePageRange(self.n1.pageRange, self.n2.pageRange)

    def compareVolumaison(self) :
        self.validation_dict["vol"] = compareVolumaison(
            self.n1.issue, self.n2.issue, self.n1.volume, self.n2.volume
        )

    def comparePublisher(self) :
        self.validation_dict["source"] = comparePublisher(
            self.n1.eissn, self.n2.eissn, self.n1.issn, self.n1.issn,
            self.n1.meeting, self.n2.meeting, self.n1.journal, self.n2.journal,
            self.n1.publiSource, self.n2.publiSource,
            self.n1.eisbn, self.n2.eisbn, self.n1.isbn, self.n2.isbn
        )


    def compareTitle(self) :
        self.validation_dict["title"] = compareTitle(
            self.n1.titleDict, self.n2.titleDict, self.threshold
        )


    def makeDecision(self) :
        """
        Make a decision based with added some cases. 
        setting 
        """
        try :
            tup = tuple(self.validation_dict.values())

            if tup[0] == 1 and tup[4] == 1 and (self.n1.typeConditor == "Thèse" or self.n2.typeConditor == "Thèse" ):
                self.result, self.duplicate_rules = (1, "1id, 1title, typeConditor=thèse")
            
            elif typeConditorCategory[self.n1.typeConditor] != typeConditorCategory[self.n2.typeConditor] :
                if (typeConditorCategory[self.n1.typeConditor] or typeConditorCategory[self.n2.typeConditor]) != "autre" :
                    self.result, self.duplicate_rules = decisionGrid[tup]
                else :
                    self.result, self.duplicate_rules = (-1, "#typeConditor")
            else :
                self.result, self.duplicate_rules = decisionGrid[tup]

        except :
            self.result, self.duplicate_rules = 99, "Record not available"

    def run(self) :
        """
        Compute all step that produce a quintuple
        """
        self.compareId()
        self.comparePageRange()
        self.compareVolumaison()
        self.comparePublisher()
        self.compareTitle()
        self.makeDecision()
        self.status = True

    def decision(self) :
        '''
        Map a quintuple to a decision according to
        the decision grid (params.py).
        '''
        if self.status :
            try :
                tup = tuple(self.validation_dict.values())
                return decisionGrid[tup]
            except :
                pass
        else :
            self.run()
            self.decision()
    
    # For Corhal records comment this two line below
    class Record(ConditorRecord) : 
        pass


#########################################
#### features comparison for Corhal #####
#########################################

class CorhalRecordComparison :
    """
    Compare two corhal records with a threshold for title distance 
    (levenshtein distance)
    """

    def __init__(self, record1, record2, threshold = 0.2) :

        self.n1 = self.Record(record1)
        self.n2 = self.Record(record2)

        self.threshold = threshold
        self.validation_dict = {}
        self.result = None
        self.duplicate_rules = None
        self.status = False

    def compareId(self):
       self.validation_dict["id"] = compareId(self.n1.doi, self.n2.doi,\
                           self.n1.nnt, self.n2.nnt,\
                           self.n1.pmId, self.n2.pmId)

    def comparePageRange(self) :
       self.validation_dict["page"] = comparePageRange(self.n1.pageRange, self.n2.pageRange)

    def compareVolumaison(self) :
        self.validation_dict["vol"] = compareVolumaison(
            self.n1.issue, self.n2.issue, self.n1.volume, self.n2.volume
        )

    def comparePublisher(self) :
        self.validation_dict["source"] = comparePublisher(
            self.n1.eissn, self.n2.eissn, self.n1.issn, self.n1.issn,
            self.n1.meeting, self.n2.meeting, self.n1.journal, self.n2.journal,
            self.n1.publiSource, self.n2.publiSource,
            self.n1.eisbn, self.n2.eisbn, self.n1.isbn, self.n2.isbn
        )


    def compareTitle(self) :
        self.validation_dict["title"] = compareTitle(
            self.n1.titleDict, self.n2.titleDict, self.threshold
        )


    def makeDecision(self) :
        """
        Make a decision based with added some cases. 
        setting 
        """
        try :
            tup = tuple(self.validation_dict.values())

            # For corhal records decomment this line below
            if tup[0] == 1 and tup[4] == 1 and (self.n1.duplicateGenre == "Thèse" or self.n2.duplicateGenre == "Thèse" ):
                self.result, self.duplicate_rules = (1, "1id, 1title, duplicateGenre=thèse")

            elif typeConditorCategory[self.n1.duplicateGenre] != typeConditorCategory[self.n2.duplicateGenre] :
                if (typeConditorCategory[self.n1.duplicateGenre] or typeConditorCategory[self.n2.duplicateGenre]) != "autre" :
                    self.result, self.duplicate_rules = decisionGrid[tup]
                else :
                    self.result, self.duplicate_rules = (-1, "#duplicateGenre")
            else :
                self.result, self.duplicate_rules = decisionGrid[tup]
        except :
            self.result, self.duplicate_rules = 99, "Record not available"

    def run(self) :
        """
        Compute all step that produce a quintuple
        """
        self.compareId()
        self.comparePageRange()
        self.compareVolumaison()
        self.comparePublisher()
        self.compareTitle()
        self.makeDecision()
        self.status = True

    def decision(self) :
        '''
        Map a quintuple to a decision according to
        the decision grid (params.py).
        '''
        if self.status :
            try :
                tup = tuple(self.validation_dict.values())
                return decisionGrid[tup]
            except :
                pass
        else :
            self.run()
            self.decision()

    ## For Corhal records decomment this two line below
    class Record(CorhalRecord) : 
       pass
