import re
import math
import string
import unidecode
import base64
import itertools as it
from unicodedata import normalize
import xml.etree.ElementTree as ET
from .params import decisionGrid, typeConditorCategory, titleStopwords

class ConditorRecord :

    def __init__(self, record) :
        assert isinstance(record, dict), "Input must be a dict"
        # id
        self.doi = record["doi"] if "doi" in record else None
        self.pmId = record["pmId"] if "pmId" in record else None
        self.nnt = record["nnt"] if "nnt" in record else None

        # Content
        self.titleDefault = record["title"]['default'] if "title" in record else None
        self.en = record["title"]["en"] if "title" in record else None
        self.fr = record["title"]["fr"] if "title" in record else None
        self.titleDict = {"default" : self.titleDefault, "en" : self.en, "fr" : self.fr}

        # container
        self.meeting = record["title"]["meeting"] if "title" in record else None
        self.journal = record["title"]["journal"] if "title" in record else None
        self.issn = record["issn"] if "issn" in record else None
        self.eissn = record["eissn"] if "eissn" in record else None
        self.eisbn = record["eisbn"] if "eisbn" in record else None
        self.isbn = record["isbn"] if "isbn" in record else None
        self.monography = record["title"]["monography"] if "title" in record else None
        self.source = record["source"] if "source" in record else None
        self.docType = record["documentType"][0] if "documentType" in record else None
        self.publiSource = getSettlement(record["teiBlob"]) if "teiBlob" in record else None

        # Positionning
        self.publiDate = record["publicationDate"] if "publicationDate" in record else None
        self.pageRange = record["pageRange"] if "pageRange" in record else None
        self.issue = record["issue"] if "issue" in record else None
        self.volume = record["volume"] if "volume" in record else None

        # Others
        self.typeConditor = record["typeConditor"] if "typeConditor" in record else None
        self.idConditor = record["idConditor"] if "idConditor" in record else None
        self.sourceUid = record['sourceUid'] if "sourceUid" in record else None

#########################################
#### Field extraction for corhal #########
#########################################

class CorhalRecord :

    def __init__(self, record) :
        assert isinstance(record, dict), "Input must be a dict"

        # id
        self.doi = record["doi"] if "doi" in record else None
        self.pmId = record["pmId"] if "pmId" in record else None
        self.nnt = record["nnt"] if "nnt" in record else None

        # Content
        self.titleDefault = record["title"]['default'] if "title" in record else None
        self.en = record["title"]["en"] if "title" in record else None
        self.fr = record["title"]["fr"] if "title" in record else None
        self.titleDict = {"default" : self.titleDefault, "en" : self.en, "fr" : self.fr}

        # container
        self.meeting = record["host"]["conference"]["name"] if "host" in record else None
        self.journal = record["host"]["title"] if "host" in record else None
        self.issn = record["host"]["issn"] if "host" in record else None
        self.eissn = record["host"]["eissn"] if "host" in record else None
        self.eisbn = record["host"]["eisbn"] if "host" in record else None
        self.isbn = record["host"]["isbn"] if "host" in record else None
        self.monography = "" #record["title"]["monography"] if "title" in record else None
        self.source = record["source"] if "source" in record else None
        self.originalGenre = record["originalGenre"] if "originalGenre" in record else None # change
        self.publiSource = getSettlement(record["teiBlob"]) if "teiBlob" in record else None

        # Positionning
        self.publiDate = record["_business"]["xPublicationDate"] if "_business" in record else None
        self.pageRange = record["host"]["page"]["range"] if "host" in record else None
        self.issue = record["host"]["issue"] if "host" in record else None
        self.volume = record["host"]["volume"] if "host" in record else None

        # Others
        self.duplicateGenre = record["_business"]["duplicateGenre"] if "_business" in record else None # change
        self.idIstex = record["idIstex"] if "idIstex" in record else None # change
        self.sourceUid = record['sourceUid'] if "sourceUid" in record else None


def getSettlement(teiblob) :
    """
    Get settlement if possible for teiblob field
    """
    decodeTeiblob = base64.b64decode(teiblob).decode('utf8')
    root = ET.fromstring(decodeTeiblob)
    dictionary = {}

    for child in root.iter("{http://www.tei-c.org/ns/1.0}meeting") :
        dictionary.setdefault("title", None)
        dictionary.setdefault("begin", None)
        dictionary.setdefault("settlement", None)
        try :
            dictionary['title'] = next(child.iter("{http://www.tei-c.org/ns/1.0}title", )).text
        except :
            pass

        try :
            dictionary["begin"] = next(child.iter("{http://www.tei-c.org/ns/1.0}date")).text
        except :
            pass

        try :
            dictionary["settlement"] = next(child.iter("{http://www.tei-c.org/ns/1.0}settlement")).text
        except :
            pass

    return dictionary



def getSameBeginSequence(string1, string2) :
    return ''.join(el[0] for el in it.takewhile(lambda t: t[0] == t[1], zip(string1, string2)))

def normalized(word) :
    if not isinstance(word, str) :
        word = str(word)

    word = word.lower().strip().translate(str.maketrans('', '', string.punctuation)) # Remove punctation
    word = normalize("NFKD", word) 
    word = unidecode.unidecode(word)
    return word



def damerauLevenshtein(seq1, seq2) :

    """Calculate the Damerau-Levenshtein distance between sequences"""
    if not isinstance(seq1, str) and not isinstance(seq1, str) :
        seq1 = str(seq1); seq2 = str(seq2)

    oneago = None
    thisrow = list(range(1, len(seq2) + 1)) + [0]
    for x in range(len(seq1)):
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in range(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)

            # This block deals with transpositions
            if (x > 0 and y > 0 and seq1[x] == seq2[y - 1]
                and seq1[x-1] == seq2[y] and seq1[x] != seq2[y]):
                thisrow[y] = min(thisrow[y], twoago[y - 2] + 1)
    return thisrow[len(seq2) - 1]

def sequenceDistance(seq1 , seq2) :
    """
    Compute relative edit distance from 2 sequences
    """
    seq1 = normalized(seq1)
    seq2 = normalized(seq2)

    maxlen = max(len(seq1), len(seq2))

    if maxlen == 0 :
        maxlen = 1

    dist = damerauLevenshtein(seq1, seq2)
    return dist/maxlen

def check(x1, x2) :
    """
    Allows to compare 2 objects : maybe string int bool list or tuple
    return 1 if they are matching, 0 if not, -1 if there are there missing values
    """
    if isinstance(x1, str) and isinstance(x2, str) :
        if x1 == "" or x2 == "" :
            return 0

        if (x1.lower().strip() in x2.lower().strip()) or\
            (x2.lower().strip() in x1.lower().strip()):
            return 1

        elif not x1 or not x2 :
            return 0

        else :
            return -1

    elif isinstance(x1, (int, bool)) and isinstance(x2, (int, bool)) :
        if x1==x2 :
            return 1
        else :
            return -1

    elif isinstance(x1, float) and isinstance(x2, float) :
        if (math.isnan(x1)== 1 or math.isnan(x2)== 1 ):
            return 0
        else :
            if x1 == x2 :
                return 1
            else :
                return -1

    elif isinstance(x1, (list, tuple)) and isinstance(x2, (list, tuple)) :
        if not x1 or not x2 :
            return 0
        else :
            return check(x1[0], x2[0])

    else : # 1 one for missing value :
        return 0

def checkPageRange(x1, x2) :
    """Compare page range """

    if not x1 or not x2 or x1 == "np" or x2 == "np":
        return 0 # Missing value

    elif isinstance(x1, float) and isinstance(x2, float):
        if math.isnan(x1) or math.isnan(x2) :
            return 0 # nan value
        else :
            if x1!=x2 :
                return -1 # Different float
            else :
                return 1 # floats not missing value which are same
    elif x1==x2 :
        return 1
    else :
        x1 = "".join([x.lower().strip() for x in str(x1).lower().strip().split("-")])
        x2 = "".join([x.lower().strip() for x in str(x2).lower().strip().split("-")])

        if x1 == x2 :
            return 1
        else : # Other cases
            if len(x2)>len(x1) :
                temp = x1
                x1 = x2
                x2 = temp

            sameSequenceAtBegining = getSameBeginSequence(x1,x2)
            x11 = x1[len(sameSequenceAtBegining):]
            x22 = x2[len(sameSequenceAtBegining):]

            diff = len(x11) - len(x22)
            begin = sameSequenceAtBegining[:diff]

            if x11.endswith(x22) and sameSequenceAtBegining.startswith(begin)\
                and 2*len(sameSequenceAtBegining) == len(x1):
                return 1
            else :
                return -1

def getrecordFromSourceUid(sourceUid, indexor):
    """
    Get a record from its sourceUdi using an indexor
    """
    if not isinstance(sourceUid, str) :
        sourceUid = str(sourceUid)
    try :
        return indexor[sourceUid]
    except Exception as err:
        return err

#####################################
# 2 We define compare functions
#####################################


def comparePublicationSource(publiSource1,publiSource2) :

    if not publiSource1 or not publiSource2 :
        return 0

    else :
        d = []
        try :
            beginDate1 = str(publiSource1['begin']).lower().strip()
            beginDate2 = str(publiSource2['begin']).lower().strip()
        except ValueError as err :
            raise err

        try :
            title1 = str(publiSource1['title']).lower().strip()
            title2 = str(publiSource2['title']).lower().strip()
        except ValueError as err :
            raise err

        try :
            settlement1 = str(publiSource1['settlement']).lower().strip()
            settlement2 = str(publiSource2['settlement']).lower().strip()
        except ValueError as err :
            raise err

        d.append(check(beginDate1, beginDate2))
        d.append(check(settlement1, settlement2))
        d.append(check(title1, title2))

        if sum(d)>=1 :
            return 1

        elif -1 in d :
            return -1

        return 0

def compareIssn(issn1, issn2) :

    if isinstance(issn1, str) and isinstance(issn2, str) :
        issn1 = "".join(issn1.strip().lower().split("-"))
        issn2 = "".join(issn2.strip().lower().split("-"))
    else :
        return 0
    return check(issn1, issn2)


def compareId(doi1, doi2, nnt1, nnt2, pmId1, pmId2) :
    """
    Compare records identifier
    """

    if doi1 and doi2 :
        if isinstance(doi1, str) and isinstance(doi2, str) :
            if str(doi1).lower().strip() == str(doi2).lower().strip() :
                return 1
            else :
                return -1

    elif nnt1 and nnt2 :
        if isinstance(nnt1, str) and isinstance(nnt2, str) :
            if str(nnt1).lower().strip() == str(nnt2).lower().strip() :
                return 1
            else :
                return -1

    elif pmId1 and pmId2 :
        if isinstance(pmId1, str) and isinstance(pmId2, str) :
            if str(pmId1).lower().strip() == str(pmId2).lower().strip() :
                return 1
            else :
                return -1

    else :
        return 0

def comparePageRange(pageRange1,pageRange2) :
    """
    Compare pageRange
    """
    return checkPageRange(pageRange1, pageRange2)

def compareVolumaison(issue1, issue2, volume1, volume2) :
    """
    Compare volumaison
    """
    if issue1 and issue2 :
        if issue1== issue2 :
            return 1
        else :
            return -1

    elif volume1 and volume2 :
        if volume1 == volume2 :
            return 1
        else :
            return -1
    else :
        return 0

def comparePublisher(eissn1, eissn2, issn1, issn2, \
                       meeting1, meeting2, journal1, journal2,\
                       publiSource1, publiSource2,eisbn1, eisbn2, isbn1, isbn2) :
    """
    Compare publisher accornding to some record fields
    """
    eissn = compareIssn(eissn1, eissn2)
    issn = compareIssn(issn1, issn2)
    meeting = check(meeting1, meeting2)
    journal = check(journal1, journal2)
    settlement = comparePublicationSource(publiSource1, publiSource2)
    isbn = check(isbn1, isbn2)
    eisbn = check(eisbn1, eisbn2)

    if settlement or issn or eissn or meeting or journal or eisbn or isbn == 1 :
        return 1
    elif settlement or issn or eissn or meeting or journal or eisbn or isbn == -1:
        return -1
    else  :
        return 0

def compareTitle(titleDict1, titleDict2, threshold = .2) :
    """
    Compare title from title subfields which are : default, en, fr.
    :param titleDict : dictionary oo key value where key are 
    "en", "fr", "default" and values are titles.
    :return -1 (different), 0 (n/a) or 1 (identical)  
    """
    
    sortedKeys =  sorted(list(titleDict1.keys()))
    # dict_keys = list(titleDict1.keys())
    # dict1_values = list(titleDict2.values())
    # dict2_values = list(titleDict2.values())
    # myrange = list(it.product(range(3),range(3)))
    # for range_ in myrange : 
    #     #print(range_)
    #     sent1 = str(dict1_values[range_[0]])
    #     sent2 = str(dict2_values[range_[1]])
    #     if len(sent1.split()) == 0 or len(sent2.split()) == 0 :
    #         return 0
    #     dist = sequenceDistance(sent1, sent2)
    #     if dist <= threshold :
    #         return 1

    #     split1 = [token for token in re.findall("\w+", sent1) if token not in titleStopwords]
    #     split2 = [token for token in re.findall("\w+", sent2) if token not in titleStopwords]
        
    #     if len(split1) == 0 or len(split2) == 0 :
    #         #and (split1[0] in titleStopwords or titleDict2Tokens[0] in titleStopwords):
    #         return 0
    
    # return -1        
    
    for key in sortedKeys :
        for otherKey in sortedKeys :
            titleDict1Tokens = titleDict1[key].split()
            titleDict2Tokens = titleDict2[otherKey].split()
            if len(titleDict1Tokens) == 0 or len(titleDict2Tokens) == 0 :
                return 0

            if len(titleDict1Tokens) == 1 or len(titleDict2Tokens) == 1 \
                and (titleDict1Tokens[0] in titleStopwords or titleDict2Tokens[0] in titleStopwords):
                return 0

            dist = sequenceDistance(titleDict1[key], titleDict2[otherKey])
            if dist <= threshold :
                return 1
    return -1

class RecordComparison :
    """
    Compare two records with a threshold for title distance 
    (levenshtein distance)
    """

    def __init__(self, record1, record2, threshold = 0.2) :

        self.n1 = self.Record(record1)
        self.n2 = self.Record(record2)

        self.threshold = threshold
        self.validation_dict = {}
        self.result = None
        self.duplicate_rules = None
        self.status = False

    def compareId(self):
       self.validation_dict["id"] = compareId(self.n1.doi, self.n2.doi,\
                           self.n1.nnt, self.n2.nnt,\
                           self.n1.pmId, self.n2.pmId)

    def comparePageRange(self) :
       self.validation_dict["page"] = comparePageRange(self.n1.pageRange, self.n2.pageRange)

    def compareVolumaison(self) :
        self.validation_dict["vol"] = compareVolumaison(
            self.n1.issue, self.n2.issue, self.n1.volume, self.n2.volume
        )

    def comparePublisher(self) :
        self.validation_dict["source"] = comparePublisher(
            self.n1.eissn, self.n2.eissn, self.n1.issn, self.n1.issn,
            self.n1.meeting, self.n2.meeting, self.n1.journal, self.n2.journal,
            self.n1.publiSource, self.n2.publiSource,
            self.n1.eisbn, self.n2.eisbn, self.n1.isbn, self.n2.isbn
        )


    def compareTitle(self) :
        self.validation_dict["title"] = compareTitle(
            self.n1.titleDict, self.n2.titleDict, self.threshold
        )


    def makeDecision(self) :
        """
        Make a decision based with added some cases. 
        setting 
        """
        try :
            tup = tuple(self.validation_dict.values())

            # Comment lines below if using corhal records
            if tup[0] == 1 and tup[4] == 1 and (self.n1.typeConditor == "Thèse" or self.n2.typeConditor == "Thèse" ):
                self.result, self.duplicate_rules = (1, "1id, 1title, typeConditor=thèse")
            
            elif typeConditorCategory[self.n1.typeConditor] != typeConditorCategory[self.n2.typeConditor] :
                if (typeConditorCategory[self.n1.typeConditor] or typeConditorCategory[self.n2.typeConditor]) != "autre" :
                    self.result, self.duplicate_rules = decisionGrid[tup]
                else :
                    self.result, self.duplicate_rules = (-1, "#duplicateGenre")
            else :
                self.result, self.duplicate_rules = decisionGrid[tup]

            ## For corhal records decomment this line below
            # if tup[0] == 1 and tup[4] == 1 and (self.n1.duplicateGenre == "Thèse" or self.n2.duplicateGenre == "Thèse" ):
                #self.result, self.duplicate_rules = (1, "1id, 1title, duplicateGenre=thèse")

            # elif typeConditorCategory[self.n1.duplicateGenre] != typeConditorCategory[self.n2.duplicateGenre] :
            #     if (typeConditorCategory[self.n1.duplicateGenre] or typeConditorCategory[self.n2.duplicateGenre]) != "autre" :
            #         self.result, self.duplicate_rules = decisionGrid[tup]
            #     else :
            #         self.result, self.duplicate_rules = (-1, "#duplicateGenre")
            # else :
            #     self.result, self.duplicate_rules = decisionGrid[tup]
        except :
            self.result, self.duplicate_rules = 99, "Record not available"

    def run(self) :
        """
        Compute all step that produce a quintuple
        """
        self.compareId()
        self.comparePageRange()
        self.compareVolumaison()
        self.comparePublisher()
        self.compareTitle()
        self.makeDecision()
        self.status = True

    def decision(self) :
        '''
        Map a quintuple to a decision according to
        the decision grid (params.py).
        '''
        if self.status :
            try :
                tup = tuple(self.validation_dict.values())
                return decisionGrid[tup]
            except :
                pass
        else :
            self.run()
            self.decision()
    
    # For Corhal records comment this two line below
    class Record(ConditorRecord) : 
        pass

    ## For Corhal records decomment this two line below
    # class Record(CorhalRecord) : 
    #    pass


