import re
import math
import string
import unidecode
import base64
import itertools as it
from unicodedata import normalize
import xml.etree.ElementTree as ET

def getSameBeginSequence(string1, string2) :
    return ''.join(el[0] for el in it.takewhile(lambda t: t[0] == t[1], zip(string1, string2)))


def getSettlement(teiblob) :
    """
    Get settlement if possible for teiblob field
    """
    decodeTeiblob = base64.b64decode(teiblob).decode('utf8')
    root = ET.fromstring(decodeTeiblob)
    dictionary = {}

    for child in root.iter("{http://www.tei-c.org/ns/1.0}meeting") :
        dictionary.setdefault("title", None)
        dictionary.setdefault("begin", None)
        dictionary.setdefault("settlement", None)
        try :
            dictionary['title'] = next(child.iter("{http://www.tei-c.org/ns/1.0}title", )).text
        except :
            pass

        try :
            dictionary["begin"] = next(child.iter("{http://www.tei-c.org/ns/1.0}date")).text
        except :
            pass

        try :
            dictionary["settlement"] = next(child.iter("{http://www.tei-c.org/ns/1.0}settlement")).text
        except :
            pass

    return dictionary


def normalized(word) :
    if not isinstance(word, str) :
        word = str(word)

    word = word.lower().strip().translate(str.maketrans('', '', string.punctuation)) # Remove punctation
    word = normalize("NFKD", word) 
    word = unidecode.unidecode(word)
    return word


def damerauLevenshtein(seq1, seq2) :

    """Calculate the Damerau-Levenshtein distance between sequences"""
    if not isinstance(seq1, str) and not isinstance(seq1, str) :
        seq1 = str(seq1); seq2 = str(seq2)

    oneago = None
    thisrow = list(range(1, len(seq2) + 1)) + [0]
    for x in range(len(seq1)):
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in range(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)

            # This block deals with transpositions
            if (x > 0 and y > 0 and seq1[x] == seq2[y - 1]
                and seq1[x-1] == seq2[y] and seq1[x] != seq2[y]):
                thisrow[y] = min(thisrow[y], twoago[y - 2] + 1)
    return thisrow[len(seq2) - 1]


def sequenceDistance(seq1 , seq2) :
    """
    Compute relative edit distance from 2 sequences
    """
    seq1 = normalized(seq1)
    seq2 = normalized(seq2)

    maxlen = max(len(seq1), len(seq2))

    if maxlen == 0 :
        maxlen = 1

    dist = damerauLevenshtein(seq1, seq2)
    return dist/maxlen

def check(x1, x2) :
    """
    Allows to compare 2 objects : maybe string int bool list or tuple
    return 1 if they are matching, 0 if not, -1 if there are there missing values
    """
    if isinstance(x1, str) and isinstance(x2, str) :
        if x1 == "" or x2 == "" :
            return 0

        if (x1.lower().strip() in x2.lower().strip()) or\
            (x2.lower().strip() in x1.lower().strip()):
            return 1

        elif not x1 or not x2 :
            return 0

        else :
            return -1

    elif isinstance(x1, (int, bool)) and isinstance(x2, (int, bool)) :
        if x1==x2 :
            return 1
        else :
            return -1

    elif isinstance(x1, float) and isinstance(x2, float) :
        if (math.isnan(x1)== 1 or math.isnan(x2)== 1 ):
            return 0
        else :
            if x1 == x2 :
                return 1
            else :
                return -1

    elif isinstance(x1, (list, tuple)) and isinstance(x2, (list, tuple)) :
        if not x1 or not x2 :
            return 0
        else :
            return check(x1[0], x2[0])

    else : # 1 one for missing value :
        return 0


def checkPageRange(x1, x2) :
    """Compare page range """

    if not x1 or not x2 or x1 == "np" or x2 == "np":
        return 0 # Missing value

    elif isinstance(x1, float) and isinstance(x2, float):
        if math.isnan(x1) or math.isnan(x2) :
            return 0 # nan value
        else :
            if x1!=x2 :
                return -1 # Different float
            else :
                return 1 # floats not missing value which are same
    elif x1==x2 :
        return 1
    else :
        x1 = "".join([x.lower().strip() for x in str(x1).lower().strip().split("-")])
        x2 = "".join([x.lower().strip() for x in str(x2).lower().strip().split("-")])

        if x1 == x2 :
            return 1
        else : # Other cases
            if len(x2)>len(x1) :
                temp = x1
                x1 = x2
                x2 = temp

            sameSequenceAtBegining = getSameBeginSequence(x1,x2)
            x11 = x1[len(sameSequenceAtBegining):]
            x22 = x2[len(sameSequenceAtBegining):]

            diff = len(x11) - len(x22)
            begin = sameSequenceAtBegining[:diff]

            if x11.endswith(x22) and sameSequenceAtBegining.startswith(begin)\
                and 2*len(sameSequenceAtBegining) == len(x1):
                return 1
            else :
                return -1

def getrecordFromSourceUid(sourceUid, indexor):
    """
    Get a record from its sourceUdi using an indexor
    """
    if not isinstance(sourceUid, str) :
        sourceUid = str(sourceUid)
    try :
        return indexor[sourceUid]
    except Exception as err:
        return err




